<!DOCTYPE html>
<html lang="en">
  <head>
    <title></title>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1, viewport-fit=cover">
    <link rel="stylesheet" href="@/static/slides.css">
  </head>
  <body class="shower list">
{{content}}
    <div class="progress"></div>
    <script src="@/static/shower.js"></script>
</body>
</html>
